from abc import ABC, abstractmethod
from jupyverse_api import Router


class Terminals(Router):
    pass


class TerminalServer(ABC):
    @abstractmethod
    async def serve(self, websocket, permissions):
        ...

    @abstractmethod
    def quit(self, websocket):
        ...



import tempfile
from pathlib import Path

import nbconvert
from fastapi import APIRouter, Depends
from fastapi.responses import FileResponse
from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.nbconvert import Nbconvert


class _Nbconvert(Nbconvert):
    def __init__(
        self,
        app: App,
        auth: Auth,
    ) -> None:
        super().__init__(app)

        router = APIRouter()

        @router.get("/api/nbconvert")
        async def get_nbconvert_formats():
            return {
                name: {"output_mimetype": nbconvert.exporters.get_exporter(name).output_mimetype}
                for name in nbconvert.exporters.get_export_names()
            }

        @router.get("/nbconvert/{format}/{path}")
        async def get_nbconvert_document(
            format: str,
            path: str,
            download: bool,
            user: User = Depends(auth.current_user(permissions={"nbconvert": ["read"]})),
        ):
            exporter = nbconvert.exporters.get_exporter(format)
            if download:
                media_type = "application/octet-stream"
            else:
                media_type = exporter.output_mimetype
            tmp_dir = Path(tempfile.mkdtemp())
            tmp_path = tmp_dir / (Path(path).stem + exporter().file_extension)
            with open(tmp_path, "wt") as f:
                f.write(exporter().from_filename(path)[0])
            return FileResponse(tmp_path, media_type=media_type, filename=tmp_path.name)

        self.include_router(router)



import os
import subprocess
import time
from pathlib import Path

import pytest
import requests


@pytest.fixture()
def cwd():
    return Path(__file__).parents[1]


@pytest.fixture()
def start_jupyverse(auth_mode, clear_users, cwd, unused_tcp_port):
    os.chdir(cwd)
    command_list = [
        "asphalt",
        "run",
        "config.yaml",
        "--set",
        f"component.components.auth.mode={auth_mode}",
        "--set",
        f"component.components.auth.clear_users={str(clear_users).lower()}",
        "--set",
        f"component.port={unused_tcp_port}",
    ]
    p = subprocess.Popen(command_list)
    url = f"http://127.0.0.1:{unused_tcp_port}"
    while True:
        try:
            requests.get(url)
        except requests.exceptions.ConnectionError:
            time.sleep(0.1)
        else:
            break
    yield url
    p.kill()
    p.wait()



import json
from typing import Dict, List

from fastapi import APIRouter, Depends, Query, Request, Response
from fastapi.responses import RedirectResponse
from fief_client import FiefAccessTokenInfo  # type: ignore
from jupyverse_api import Router
from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User

from .backend import Backend
from .config import _AuthFiefConfig


class _AuthFief(Backend, Auth, Router):
    def __init__(
        self,
        app: App,
        auth_fief_config: _AuthFiefConfig,
    ) -> None:
        Router.__init__(self, app)
        Backend.__init__(self, auth_fief_config)

        router = APIRouter()

        @router.get("/auth-callback", name="auth_callback")
        async def auth_callback(request: Request, response: Response, code: str = Query(...)):
            redirect_uri = str(request.url_for("auth_callback"))
            tokens, _ = await self.fief.auth_callback(code, redirect_uri)

            response = RedirectResponse(request.url_for("root"))
            response.set_cookie(
                self.SESSION_COOKIE_NAME,
                tokens["access_token"],
                max_age=tokens["expires_in"],
                httponly=True,
                secure=False,
            )

            return response

        @router.get("/api/me")
        async def get_api_me(
            request: Request,
            user: User = Depends(self.current_user()),
            access_token_info: FiefAccessTokenInfo = Depends(self.auth.authenticated()),
        ):
            checked_permissions: Dict[str, List[str]] = {}
            permissions = json.loads(
                dict(request.query_params).get("permissions", "{}").replace("'", '"')
            )
            if permissions:
                user_permissions: Dict[str, List[str]] = {}
                for permission in access_token_info["permissions"]:
                    resource, action = permission.split(":")
                    if resource not in user_permissions:
                        user_permissions[resource] = []
                    user_permissions[resource].append(action)
                for resource, actions in permissions.items():
                    user_resource_permissions = user_permissions.get(resource, [])
                    allowed = checked_permissions[resource] = []
                    for action in actions:
                        if action in user_resource_permissions:
                            allowed.append(action)

            keys = ["username", "name", "display_name", "initials", "avatar_url", "color"]
            identity = {k: getattr(user, k) for k in keys}
            return {
                "identity": identity,
                "permissions": checked_permissions,
            }

        self.include_router(router)



import asyncio
from datetime import datetime
from pathlib import Path
from typing import Optional, Tuple
from uuid import uuid4

from fastapi import (
    APIRouter,
    Depends,
    HTTPException,
    Request,
    Response,
    WebSocketDisconnect,
    status,
)
from jupyter_ydoc import ydocs as YDOCS  # type: ignore
from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.contents import Contents
from jupyverse_api.yjs import Yjs
from ypy_websocket.websocket_server import WebsocketServer, YRoom
from ypy_websocket.ystore import BaseYStore, SQLiteYStore, YDocNotFound
from ypy_websocket.yutils import YMessageType, YSyncMessageType

from .models import CreateDocumentSession

YFILE = YDOCS["file"]
AWARENESS = 1
RENAME_SESSION = 127
SERVER_SESSION = uuid4().hex


class JupyterSQLiteYStore(SQLiteYStore):
    db_path = ".jupyter_ystore.db"


class _Yjs(Yjs):
    def __init__(
        self,
        app: App,
        auth: Auth,
        contents: Contents,
    ) -> None:
        super().__init__(app)

        class YDocWebSocketHandler:
            saving_document: Optional[asyncio.Task]
            websocket_server = JupyterWebsocketServer(rooms_ready=False, auto_clean_rooms=False)

            def __init__(self, websocket, path, permissions):
                self.websocket = websocket
                self.can_write = permissions is None or "write" in permissions.get("yjs", [])
                self.lock = asyncio.Lock()
                self.room = self.websocket_server.get_room(self.websocket.path)
                self.set_file_info(path)

            async def get_file_info(self) -> Tuple[str, str, str]:
                room_name = self.websocket_server.get_room_name(self.room)
                file_format, file_type, file_id = room_name.split(":", 2)
                file_path = await contents.file_id_manager.get_path(file_id)
                if file_path is None:
                    raise RuntimeError(f"File {self.room.document.path} cannot be found anymore")
                if file_path != self.room.document.path:
                    self.room.document.path = file_path
                return file_format, file_type, file_path

            def set_file_info(self, value: str) -> None:
                self.websocket_server.rename_room(value, from_room=self.room)
                self.websocket.path = value

            async def serve(self):
                self.set_file_info(self.websocket.path)
                self.saving_document = None
                self.room.on_message = self.on_message

                # cancel the deletion of the room if it was scheduled
                if not self.room.is_transient and self.room.cleaner is not None:
                    self.room.cleaner.cancel()

                if not self.room.is_transient and not self.room.ready:
                    file_format, file_type, file_path = await self.get_file_info()
                    is_notebook = file_type == "notebook"
                    model = await contents.read_content(file_path, True, as_json=is_notebook)
                    self.last_modified = to_datetime(model.last_modified)
                    # check again if ready, because loading the file is async
                    if not self.room.ready:
                        # try to apply Y updates from the YStore for this document
                        try:
                            await self.room.ystore.apply_updates(self.room.ydoc)
                            read_from_source = False
                        except YDocNotFound:
                            # YDoc not found in the YStore, create the document from
                            # the source file (no change history)
                            read_from_source = True
                        if not read_from_source:
                            # if YStore updates and source file are out-of-sync, resync updates
                            # with source
                            if self.room.document.source != model.content:
                                read_from_source = True
                        if read_from_source:
                            self.room.document.source = model.content
                            await self.room.ystore.encode_state_as_update(self.room.ydoc)

                        self.room.document.dirty = False
                        self.room.ready = True
                        self.room.watcher = asyncio.create_task(self.watch_file())
                        # save the document when changed
                        self.room.document.observe(self.on_document_change)

                await self.websocket_server.serve(self.websocket)
                if not self.room.is_transient and not self.room.clients:
                    # no client in this room after we disconnect
                    # keep the document for a while in case someone reconnects
                    self.room.cleaner = asyncio.create_task(self.clean_room())

            async def on_message(self, message: bytes) -> bool:
                """
                Called whenever a message is received, before forwarding it to other clients.

                :param message: received message
                :returns: True if the message must be discarded, False otherwise (default: False).
                """
                skip = False
                byte = message[0]
                msg = message[1:]
                if byte == RENAME_SESSION:
                    # The client moved the document to a different location. After receiving this
                    # message, we make the current document available under a different url.
                    # The other clients are automatically notified of this change because
                    # the path is shared through the Yjs document as well.
                    new_room_name = msg.decode("utf-8")
                    self.set_file_info(new_room_name)
                    self.websocket_server.rename_room(new_room_name, from_room=self.room)
                    # send rename acknowledge
                    await self.websocket.send(bytes([RENAME_SESSION, 1]))
                elif byte == AWARENESS:
                    # changes = self.room.awareness.get_changes(msg)
                    # # filter out message depending on changes
                    # skip = True
                    pass
                elif byte == YMessageType.SYNC:
                    if not self.can_write and msg[0] == YSyncMessageType.SYNC_UPDATE:
                        skip = True
                else:
                    skip = True
                return skip

            async def watch_file(self):
                file_format, file_type, file_path = await self.get_file_info()
                while True:
                    watcher = contents.file_id_manager.watch(file_path)
                    async for changes in watcher:
                        file_format, file_type, new_file_path = await self.get_file_info()
                        if new_file_path != file_path:
                            # file was renamed
                            contents.file_id_manager.unwatch(file_path, watcher)
                            file_path = new_file_path
                            break
                        await self.maybe_load_document()

            async def maybe_load_document(self):
                file_format, file_type, file_path = await self.get_file_info()
                async with self.lock:
                    model = await contents.read_content(file_path, False)
                # do nothing if the file was saved by us
                if self.last_modified < to_datetime(model.last_modified):
                    is_notebook = file_type == "notebook"
                    model = await contents.read_content(file_path, True, as_json=is_notebook)
                    self.room.document.source = model.content
                    self.last_modified = to_datetime(model.last_modified)

            async def clean_room(self) -> None:
                await asyncio.sleep(60)  # FIXME: pass in config
                if self.room.watcher:
                    self.room.watcher.cancel()
                self.room.document.unobserve()
                self.websocket_server.delete_room(room=self.room)

            def on_document_change(self, target, event):
                if target == "state" and "dirty" in event.keys:
                    dirty = event.keys["dirty"]["newValue"]
                    if not dirty:
                        # we cleared the dirty flag, nothing to save
                        return
                # unobserve and observe again because the structure of the document may have changed
                # e.g. a new cell added to a notebook
                self.room.document.unobserve()
                self.room.document.observe(self.on_document_change)
                if self.saving_document is not None and not self.saving_document.done():
                    # the document is being saved, cancel that
                    self.saving_document.cancel()
                    self.saving_document = None
                self.saving_document = asyncio.create_task(self.maybe_save_document())

            async def maybe_save_document(self):
                # save after 1 second of inactivity to prevent too frequent saving
                await asyncio.sleep(1)
                # if the room cannot be found, don't save
                try:
                    file_format, file_type, file_path = await self.get_file_info()
                except Exception:
                    return
                is_notebook = file_type == "notebook"
                async with self.lock:
                    model = await contents.read_content(file_path, True, as_json=is_notebook)
                if self.last_modified < to_datetime(model.last_modified):
                    # file changed on disk, let's revert
                    self.room.document.source = model.content
                    self.last_modified = to_datetime(model.last_modified)
                    return
                if model.content != self.room.document.source:
                    # don't save if not needed
                    # this also prevents the dirty flag from bouncing between windows of
                    # the same document opened as different types (e.g. notebook/text editor)
                    format = "json" if file_type == "notebook" else "text"
                    content = {
                        "content": self.room.document.source,
                        "format": format,
                        "path": file_path,
                        "type": file_type,
                    }
                    async with self.lock:
                        await contents.write_content(content)
                        model = await contents.read_content(file_path, False)
                        self.last_modified = to_datetime(model.last_modified)
                self.room.document.dirty = False

        router = APIRouter()

        async def serve_room(websocket_permissions, path):
            if websocket_permissions is None:
                return
            websocket, permissions = websocket_permissions
            await websocket.accept()
            socket = YDocWebSocketHandler(WebsocketAdapter(websocket, path), path, permissions)
            await socket.serve()

        @router.websocket("/api/yjs/{path:path}")
        async def yjs_websocket(
            path,
            websocket_permissions=Depends(
                auth.websocket_auth(permissions={"yjs": ["read", "write"]})
            ),
        ):
            await serve_room(websocket_permissions, path)

        @router.websocket("/api/collaboration/room/{path:path}")
        async def collaboration_room_websocket(
            path,
            websocket_permissions=Depends(
                auth.websocket_auth(permissions={"yjs": ["read", "write"]})
            ),
        ):
            await serve_room(websocket_permissions, path)

        @router.put("/api/collaboration/session/{path:path}", status_code=200)
        async def create_roomid(
            path,
            request: Request,
            response: Response,
            user: User = Depends(auth.current_user(permissions={"contents": ["read"]})),
        ):
            # we need to process the request manually
            # see https://github.com/tiangolo/fastapi/issues/3373#issuecomment-1306003451
            create_document_session = CreateDocumentSession(**(await request.json()))
            idx = await contents.file_id_manager.get_id(path)
            if idx is not None:
                return {
                    "format": create_document_session.format,
                    "type": create_document_session.type,
                    "fileId": idx,
                    "sessionId": SERVER_SESSION,
                }

            idx = await contents.file_id_manager.index(path)
            if idx is None:
                raise HTTPException(status_code=404, detail=f"File {path} does not exist")

            response.status_code = status.HTTP_201_CREATED
            return {
                "format": create_document_session.format,
                "type": create_document_session.type,
                "fileId": idx,
                "sessionId": SERVER_SESSION,
            }

        self.include_router(router)

        self.YDocWebSocketHandler = YDocWebSocketHandler


def to_datetime(iso_date: str) -> datetime:
    return datetime.fromisoformat(iso_date.rstrip("Z"))


class WebsocketAdapter:
    """An adapter to make a Starlette's WebSocket look like a ypy-websocket's WebSocket"""

    def __init__(self, websocket, path: str):
        self._websocket = websocket
        self._path = path

    @property
    def path(self) -> str:
        return self._path

    @path.setter
    def path(self, value: str) -> None:
        self._path = value

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            message = await self._websocket.receive_bytes()
        except WebSocketDisconnect:
            raise StopAsyncIteration()
        return message

    async def send(self, message):
        await self._websocket.send_bytes(message)

    async def recv(self):
        return await self._websocket.receive_bytes()


class DocumentRoom(YRoom):
    """A Y room for a possibly stored document (e.g. a notebook)."""

    is_transient = False

    def __init__(self, type: str, ystore: BaseYStore):
        super().__init__(ready=False, ystore=ystore)
        self.type = type
        self.cleaner = None
        self.watcher = None
        self.document = YDOCS.get(type, YFILE)(self.ydoc)


class TransientRoom(YRoom):
    """A Y room for sharing state (e.g. awareness)."""

    is_transient = True


class JupyterWebsocketServer(WebsocketServer):
    def get_room(self, path: str) -> YRoom:
        if path not in self.rooms.keys():
            if path.count(":") >= 2:
                # it is a stored document (e.g. a notebook)
                file_format, file_type, file_path = path.split(":", 2)
                p = Path(file_path)
                updates_file_path = (p.parent / f".{file_type}:{p.name}.y").as_posix()
                ystore = JupyterSQLiteYStore(path=updates_file_path)  # FIXME: pass in config
                self.rooms[path] = DocumentRoom(file_type, ystore)
            else:
                # it is a transient document (e.g. awareness)
                self.rooms[path] = TransientRoom()
        return self.rooms[path]



import json
import logging
import os
import sys
from glob import glob
from http import HTTPStatus
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import json5  # type: ignore
import pkg_resources
from babel import Locale
from fastapi import APIRouter, Depends, Response, status
from fastapi.responses import FileResponse, RedirectResponse
from fastapi.staticfiles import StaticFiles
from starlette.requests import Request

import jupyverse
from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.jupyterlab import JupyterLabConfig
from jupyverse_api.lab import Lab


logger = logging.getLogger("lab")


class _Lab(Lab):
    def __init__(
        self,
        app: App,
        auth: Auth,
        frontend_config: FrontendConfig,
        jupyterlab_config: Optional[JupyterLabConfig],
    ) -> None:
        super().__init__(app)

        self.auth = auth
        self.frontend_config = frontend_config

        if jupyterlab_config is not None:
            import jupyterlab as jupyterlab_module  # type: ignore

            jlab_dev_mode = jupyterlab_config.dev_mode
        else:
            jlab_dev_mode = False

        self.locale = "en"

        self.prefix_dir = Path(sys.prefix)
        if jlab_dev_mode:
            self.jlab_dir = Path(jupyterlab_module.__file__).parents[1] / "dev_mode"
        else:
            self.jlab_dir = self.prefix_dir / "share" / "jupyter" / "lab"

    def init_router(
        self, router: APIRouter, redirect_after_root: str
    ) -> Tuple[Path, List[Dict[str, Any]]]:
        extensions_dir = self.prefix_dir / "share" / "jupyter" / "labextensions"
        federated_extensions, disabled_extensions = self.get_federated_extensions(extensions_dir)

        for ext in federated_extensions:
            name = ext["name"]
            self.mount(
                f"/lab/extensions/{name}/static",
                StaticFiles(directory=extensions_dir / name / "static"),
                name=name,
            )

        self.mount(
            "/lab/api/themes",
            StaticFiles(directory=self.jlab_dir / "themes"),
            name="themes",
        )

        @router.get("/", name="root")
        async def get_root(
            response: Response,
            user: User = Depends(self.auth.current_user()),
        ):
            # auto redirect
            response.status_code = status.HTTP_302_FOUND
            response.headers["Location"] = self.frontend_config.base_url + redirect_after_root

        @router.get("/favicon.ico")
        async def get_favicon():
            return FileResponse(Path(jupyverse.__file__).parent / "static" / "favicon.ico")

        @router.get("/static/notebook/components/MathJax/{rest_of_path:path}")
        async def get_mathjax(rest_of_path):
            return RedirectResponse(
                "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.7/" + rest_of_path
            )

        @router.get("/lab/api/listings/@jupyterlab/extensionmanager-extension/listings.json")
        async def get_listings(user: User = Depends(self.auth.current_user())):
            return {
                "blocked_extensions_uris": [],
                "allowed_extensions_uris": [],
                "blocked_extensions": [],
                "allowed_extensions": [],
            }

        @router.get("/lab/api/extensions")
        async def get_extensions(user: User = Depends(self.auth.current_user())):
            return federated_extensions

        @router.get("/lab/api/translations/")
        async def get_translations_(
            user: User = Depends(self.auth.current_user()),
        ):
            return RedirectResponse(f"{self.frontend_config.base_url}lab/api/translations")

        @router.get("/lab/api/translations")
        async def get_translations(user: User = Depends(self.auth.current_user())):
            locale = Locale.parse("en")
            display_name = (locale.get_display_name(self.locale) or "").capitalize()
            native_name = (locale.get_display_name() or "").capitalize()
            data = {
                "en": {
                    "displayName": display_name,
                    "nativeName": native_name,
                }
            }
            for ep in pkg_resources.iter_entry_points(group="jupyterlab.languagepack"):
                locale = Locale.parse(ep.name)
                data[ep.name] = {
                    "displayName": display_name,
                    "nativeName": native_name,
                }
            return {"data": data, "message": ""}

        @router.get("/lab/api/translations/{language}")
        async def get_translation(
            language,
            user: User = Depends(self.auth.current_user()),
        ):
            if language == "en":
                self.locale = language
                return {}
            for ep in pkg_resources.iter_entry_points(group="jupyterlab.languagepack"):
                if ep.name == language:
                    break
            else:
                return {"data": {}, "message": f"Language pack '{language}' not installed!"}
            self.locale = language
            package = ep.load()
            data = {}
            for path in (Path(package.__file__).parent / "locale" / language / "LC_MESSAGES").glob(
                "*.json"
            ):
                with open(path) as f:
                    data.update({path.stem: json.load(f)})
            return {"data": data, "message": ""}

        @router.get("/lab/api/settings/{name0}/{name1}:{name2}")
        async def get_setting(
            name0,
            name1,
            name2,
            user: User = Depends(self.auth.current_user()),
        ):
            with open(self.jlab_dir / "static" / "package.json") as f:
                package = json.load(f)
            if name0 in ["@jupyterlab", "@retrolab"]:
                schemas_parent = self.jlab_dir
            else:
                schemas_parent = extensions_dir / name0 / name1
            with open(schemas_parent / "schemas" / name0 / name1 / f"{name2}.json") as f:
                schema = json.load(f)
            key = f"{name1}:{name2}"
            setting = {
                "id": f"@jupyterlab/{key}",
                "schema": schema,
                "version": package["version"],
                "raw": "{}",
                "settings": {},
                "last_modified": None,
                "created": None,
            }
            if user:
                user_settings = json.loads(user.settings)
                if key in user_settings:
                    setting.update(user_settings[key])
                    setting["settings"] = json5.loads(user_settings[key]["raw"])
            return setting

        @router.put(
            "/lab/api/settings/@jupyterlab/{name0}:{name1}",
            status_code=204,
        )
        async def change_setting(
            request: Request,
            name0,
            name1,
            user: User = Depends(self.auth.current_user()),
            user_update=Depends(self.auth.update_user),
        ):
            settings = json.loads(user.settings)
            settings[f"{name0}:{name1}"] = await request.json()
            await user_update({"settings": json.dumps(settings)})
            return Response(status_code=HTTPStatus.NO_CONTENT.value)

        @router.get("/lab/api/settings")
        async def get_settings(user: User = Depends(self.auth.current_user())):
            with open(self.jlab_dir / "static" / "package.json") as f:
                package = json.load(f)
            if user:
                user_settings = json.loads(user.settings)
            else:
                user_settings = {}
            settings = []
            for path in (self.jlab_dir / "schemas" / "@jupyterlab").glob("*/*.json"):
                with open(path) as f:
                    schema = json.load(f)
                key = f"{path.parent.name}:{path.stem}"
                setting = {
                    "id": f"@jupyterlab/{key}",
                    "schema": schema,
                    "version": package["version"],
                    "raw": "{}",
                    "settings": {},
                    "warning": None,
                    "last_modified": None,
                    "created": None,
                }
                if key in user_settings:
                    setting.update(user_settings[key])
                    setting["settings"] = json5.loads(user_settings[key]["raw"])
                settings.append(setting)
            return {"settings": settings}

        return self.prefix_dir, federated_extensions

    def get_federated_extensions(self, extensions_dir: Path) -> Tuple[List, List]:
        federated_extensions = []
        disabled_extensions = []

        for path in glob(os.path.join(extensions_dir, "**", "package.json"), recursive=True):
            with open(path) as f:
                package = json.load(f)
            if "jupyterlab" not in package:
                continue
            extension = package["jupyterlab"]["_build"]
            extension["name"] = package["name"]
            extension["description"] = package["description"]
            federated_extensions.append(extension)

            for ext in package["jupyterlab"].get("disabledExtensions", []):
                disabled_extensions.append(ext)

        return federated_extensions, disabled_extensions



import contextlib
import json
import logging
from typing import Any, Callable, Dict, List, Optional, Tuple

from fastapi import APIRouter, Depends, Request
from jupyverse_api import Router
from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.frontend import FrontendConfig
from sqlalchemy import select  # type: ignore

from .backends import get_backend
from .config import _AuthConfig
from .db import get_db

from .models import UserCreate, UserRead, UserUpdate


logger = logging.getLogger("auth")


def auth_factory(
    app: App,
    auth_config: _AuthConfig,
    frontend_config: FrontendConfig,
):
    db = get_db(auth_config)
    backend = get_backend(auth_config, frontend_config, db)

    get_async_session_context = contextlib.asynccontextmanager(db.get_async_session)
    get_user_db_context = contextlib.asynccontextmanager(db.get_user_db)
    get_user_manager_context = contextlib.asynccontextmanager(backend.get_user_manager)

    @contextlib.asynccontextmanager
    async def _get_user_manager():
        async with get_async_session_context() as session:
            async with get_user_db_context(session) as user_db:
                async with get_user_manager_context(user_db) as user_manager:
                    yield user_manager

    async def create_user(**kwargs):
        async with _get_user_manager() as user_manager:
            await user_manager.create(UserCreate(**kwargs))

    async def update_user(user, **kwargs):
        async with _get_user_manager() as user_manager:
            await user_manager.update(UserUpdate(**kwargs), user)

    async def get_user_by_email(user_email):
        async with _get_user_manager() as user_manager:
            return await user_manager.get_by_email(user_email)

    class _Auth(Auth, Router):
        def __init__(self) -> None:
            super().__init__(app)

            self.db = db

            router = APIRouter()

            @router.get("/auth/users")
            async def get_users(
                user: UserRead = Depends(backend.current_user(permissions={"admin": ["read"]})),
            ):
                async with db.async_session_maker() as session:
                    statement = select(db.User)
                    users = (await session.execute(statement)).unique().all()
                return [usr.User for usr in users if usr.User.is_active]

            @router.get("/api/me")
            async def get_api_me(
                request: Request,
                user: UserRead = Depends(backend.current_user()),
            ):
                checked_permissions: Dict[str, List[str]] = {}
                permissions = json.loads(
                    dict(request.query_params).get("permissions", "{}").replace("'", '"')
                )
                if permissions:
                    user_permissions = user.permissions
                    for resource, actions in permissions.items():
                        user_resource_permissions = user_permissions.get(resource)
                        if user_resource_permissions is None:
                            continue
                        allowed = checked_permissions[resource] = []
                        for action in actions:
                            if action in user_resource_permissions:
                                allowed.append(action)

                keys = ["username", "name", "display_name", "initials", "avatar_url", "color"]
                identity = {k: getattr(user, k) for k in keys}
                return {
                    "identity": identity,
                    "permissions": checked_permissions,
                }

            # redefine GET /me because we want our current_user dependency
            # it is first defined in users_router and so it wins over the one in
            # fapi_users.get_users_router
            users_router = APIRouter()

            @users_router.get("/me")
            async def get_me(
                user: UserRead = Depends(backend.current_user(permissions={"admin": ["read"]})),
            ):
                return user

            users_router.include_router(backend.fapi_users.get_users_router(UserRead, UserUpdate))

            # Cookie based auth login and logout
            self.include_router(
                backend.fapi_users.get_auth_router(backend.cookie_authentication), prefix="/auth"
            )
            self.include_router(
                backend.fapi_users.get_register_router(UserRead, UserCreate),
                prefix="/auth",
                dependencies=[Depends(backend.current_user(permissions={"admin": ["write"]}))],
            )
            self.include_router(users_router, prefix="/auth/user")

            # GitHub OAuth register router
            self.include_router(
                backend.fapi_users.get_oauth_router(
                    backend.github_authentication, backend.github_cookie_authentication, db.secret
                ),
                prefix="/auth/github",
            )
            self.include_router(router)

            self.create_user = create_user
            self.__update_user = update_user
            self.get_user_by_email = get_user_by_email

        async def _update_user(self, user, **kwargs):
            return await self.__update_user(user, **kwargs)

        def current_user(self, permissions: Optional[Dict[str, List[str]]] = None) -> Callable:
            return backend.current_user(permissions)

        async def update_user(self, update_user=Depends(backend.update_user)) -> Callable:
            return update_user

        def websocket_auth(
            self,
            permissions: Optional[Dict[str, List[str]]] = None,
        ) -> Callable[[], Tuple[Any, Dict[str, List[str]]]]:
            return backend.websocket_auth(permissions)

    return _Auth()



import json
import os
import shutil
from datetime import datetime
from http import HTTPStatus
from pathlib import Path
from typing import Dict, List, Optional, Union, cast

from anyio import open_file
from fastapi import APIRouter, Depends, HTTPException, Response
from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.contents import Contents, Content, SaveContent
from starlette.requests import Request

from .fileid import FileIdManager
from .models import Checkpoint, CreateContent, RenameContent


class _Contents(Contents):
    def __init__(self, app: App, auth: Auth):
        super().__init__(app=app)

        router = APIRouter()

        @router.post(
            "/api/contents/{path:path}/checkpoints",
            status_code=201,
        )
        async def create_checkpoint(
            path, user: User = Depends(auth.current_user(permissions={"contents": ["write"]}))
        ):
            src_path = Path(path)
            dst_path = Path(".ipynb_checkpoints") / f"{src_path.stem}-checkpoint{src_path.suffix}"
            try:
                dst_path.parent.mkdir(exist_ok=True)
                shutil.copyfile(src_path, dst_path)
            except Exception:
                # FIXME: return error code?
                return []
            mtime = get_file_modification_time(dst_path)
            return Checkpoint(**{"id": "checkpoint", "last_modified": mtime})

        @router.post(
            "/api/contents{path:path}",
            status_code=201,
        )
        async def create_content(
            path: Optional[str],
            request: Request,
            user: User = Depends(auth.current_user(permissions={"contents": ["write"]})),
        ):
            create_content = CreateContent(**(await request.json()))
            content_path = Path(create_content.path)
            if create_content.type == "notebook":
                available_path = get_available_path(content_path / "Untitled.ipynb")
                async with await open_file(available_path, "w") as f:
                    await f.write(
                        json.dumps(
                            {"cells": [], "metadata": {}, "nbformat": 4, "nbformat_minor": 5}
                        )
                    )
                src_path = available_path
                dst_path = (
                    Path(".ipynb_checkpoints") / f"{src_path.stem}-checkpoint{src_path.suffix}"
                )
                try:
                    dst_path.parent.mkdir(exist_ok=True)
                    shutil.copyfile(src_path, dst_path)
                except Exception:
                    # FIXME: return error code?
                    pass
            elif create_content.type == "directory":
                name = "Untitled Folder"
                available_path = get_available_path(content_path / name, sep=" ")
                available_path.mkdir(parents=True, exist_ok=True)
            else:
                assert create_content.ext is not None
                available_path = get_available_path(
                    content_path / ("untitled" + create_content.ext)
                )
                open(available_path, "w").close()

            return await self.read_content(available_path, False)

        @router.get("/api/contents")
        async def get_root_content(
            content: int,
            user: User = Depends(auth.current_user(permissions={"contents": ["read"]})),
        ):
            return await self.read_content("", bool(content))

        @router.get("/api/contents/{path:path}/checkpoints")
        async def get_checkpoint(
            path, user: User = Depends(auth.current_user(permissions={"contents": ["read"]}))
        ):
            src_path = Path(path)
            dst_path = Path(".ipynb_checkpoints") / f"{src_path.stem}-checkpoint{src_path.suffix}"
            if not dst_path.exists():
                return []
            mtime = get_file_modification_time(dst_path)
            return [Checkpoint(**{"id": "checkpoint", "last_modified": mtime})]

        @router.get("/api/contents/{path:path}")
        async def get_content(
            path: str,
            content: int = 0,
            user: User = Depends(auth.current_user(permissions={"contents": ["read"]})),
        ):
            return await self.read_content(path, bool(content))

        @router.put("/api/contents/{path:path}")
        async def save_content(
            path,
            request: Request,
            response: Response,
            user: User = Depends(auth.current_user(permissions={"contents": ["write"]})),
        ):
            content = SaveContent(**(await request.json()))
            try:
                await self.write_content(content)
            except Exception:
                raise HTTPException(status_code=404, detail=f"Error saving {content.path}")
            return await self.read_content(content.path, False)

        @router.delete(
            "/api/contents/{path:path}",
            status_code=204,
        )
        async def delete_content(
            path,
            user: User = Depends(auth.current_user(permissions={"contents": ["write"]})),
        ):
            p = Path(path)
            if p.exists():
                if p.is_dir():
                    shutil.rmtree(p)
                else:
                    p.unlink()
            return Response(status_code=HTTPStatus.NO_CONTENT.value)

        @router.patch("/api/contents/{path:path}")
        async def rename_content(
            path,
            request: Request,
            user: User = Depends(auth.current_user(permissions={"contents": ["write"]})),
        ):
            rename_content = RenameContent(**(await request.json()))
            Path(path).rename(rename_content.path)
            return await self.read_content(rename_content.path, False)

        self.include_router(router)

    async def read_content(
        self, path: Union[str, Path], get_content: bool, as_json: bool = False
    ) -> Content:
        if isinstance(path, str):
            path = Path(path)
        content: Optional[Union[str, Dict, List[Dict]]] = None
        if get_content:
            if path.is_dir():
                content = [
                    (await self.read_content(subpath, get_content=False)).dict()
                    for subpath in path.iterdir()
                    if not subpath.name.startswith(".")
                ]
            elif path.is_file() or path.is_symlink():
                try:
                    async with await open_file(path) as f:
                        content = await f.read()
                    if as_json:
                        content = json.loads(content)
                except Exception:
                    raise HTTPException(status_code=404, detail="Item not found")
        format: Optional[str] = None
        if path.is_dir():
            size = None
            type = "directory"
            format = "json"
            mimetype = None
        elif path.is_file() or path.is_symlink():
            size = get_file_size(path)
            if path.suffix == ".ipynb":
                type = "notebook"
                format = None
                mimetype = None
                if content is not None:
                    nb: dict
                    if as_json:
                        content = cast(Dict, content)
                        nb = content
                    else:
                        content = cast(str, content)
                        nb = json.loads(content)
                    for cell in nb["cells"]:
                        if "metadata" not in cell:
                            cell["metadata"] = {}
                        cell["metadata"].update({"trusted": False})
                    if not as_json:
                        content = json.dumps(nb)
            elif path.suffix == ".json":
                type = "json"
                format = "text"
                mimetype = "application/json"
            else:
                type = "file"
                format = None
                mimetype = "text/plain"
        else:
            raise HTTPException(status_code=404, detail="Item not found")

        return Content(
            **{
                "name": path.name,
                "path": path.as_posix(),
                "last_modified": get_file_modification_time(path),
                "created": get_file_creation_time(path),
                "content": content,
                "format": format,
                "mimetype": mimetype,
                "size": size,
                "writable": is_file_writable(path),
                "type": type,
            }
        )

    async def write_content(self, content: Union[SaveContent, Dict]) -> None:
        if not isinstance(content, SaveContent):
            content = SaveContent(**content)
        async with await open_file(content.path, "w") as f:
            if content.format == "json":
                dict_content = cast(Dict, content.content)
                if content.type == "notebook":
                    # see https://github.com/jupyterlab/jupyterlab/issues/11005
                    if "metadata" in dict_content and "orig_nbformat" in dict_content["metadata"]:
                        del dict_content["metadata"]["orig_nbformat"]
                await f.write(json.dumps(dict_content, indent=2))
            else:
                content.content = cast(str, content.content)
                await f.write(content.content)

    @property
    def file_id_manager(self):
        return FileIdManager()


def get_available_path(path: Path, sep: str = ""):
    directory = path.parent
    name = Path(path.name)
    i = None
    while True:
        if i is None:
            i_str = ""
            i = 1
        else:
            i_str = str(i)
            i += 1
        if i_str:
            i_str = sep + i_str
        available_path = directory / (name.stem + i_str + name.suffix)
        if not available_path.exists():
            return available_path


def get_file_modification_time(path: Path):
    if path.exists():
        return datetime.utcfromtimestamp(path.stat().st_mtime).isoformat() + "Z"


def get_file_creation_time(path: Path):
    if path.exists():
        return datetime.utcfromtimestamp(path.stat().st_ctime).isoformat() + "Z"


def get_file_size(path: Path) -> Optional[int]:
    if path.exists():
        return path.stat().st_size
    raise HTTPException(status_code=404, detail="Item not found")


def is_file_writable(path: Path) -> bool:
    if path.exists():
        if path.is_dir():
            # FIXME
            return True
        else:
            return os.access(path, os.W_OK)
    return False



from asphalt.core import Component, Context
from jupyverse_api.auth import Auth

from .backends import _NoAuth


class NoAuthComponent(Component):
    async def start(
        self,
        ctx: Context,
    ) -> None:
        no_auth = _NoAuth()
        ctx.add_resource(no_auth, types=Auth)



import asyncio
import logging
from typing import Dict, List, Optional
from uuid import uuid4

import aiosqlite
from anyio import Path
from jupyverse_api import Singleton
from watchfiles import Change, awatch

logger = logging.getLogger("contents")


class Watcher:
    def __init__(self, path: str) -> None:
        self.path = path
        self._event = asyncio.Event()

    def __aiter__(self):
        return self

    async def __anext__(self):
        await self._event.wait()
        self._event.clear()
        return self._change

    def notify(self, change):
        self._change = change
        self._event.set()


class FileIdManager(metaclass=Singleton):
    db_path: str
    initialized: asyncio.Event
    watchers: Dict[str, List[Watcher]]
    lock: asyncio.Lock

    def __init__(self, db_path: str = "fileid.db"):
        self.db_path = db_path
        self.initialized = asyncio.Event()
        self.watchers = {}
        self.watch_files_task = asyncio.create_task(self.watch_files())
        self.stop_watching_files = asyncio.Event()
        self.stopped_watching_files = asyncio.Event()
        self.lock = asyncio.Lock()

    async def get_id(self, path: str) -> Optional[str]:
        await self.initialized.wait()
        async with self.lock:
            async with aiosqlite.connect(self.db_path) as db:
                async with db.execute("SELECT id FROM fileids WHERE path = ?", (path,)) as cursor:
                    async for idx, in cursor:
                        return idx
                    return None

    async def get_path(self, idx: str) -> Optional[str]:
        await self.initialized.wait()
        async with self.lock:
            async with aiosqlite.connect(self.db_path) as db:
                async with db.execute("SELECT path FROM fileids WHERE id = ?", (idx,)) as cursor:
                    async for path, in cursor:
                        return path
                    return None

    async def index(self, path: str) -> Optional[str]:
        await self.initialized.wait()
        async with self.lock:
            async with aiosqlite.connect(self.db_path) as db:
                apath = Path(path)
                if not await apath.exists():
                    return None

                idx = uuid4().hex
                mtime = (await apath.stat()).st_mtime
                await db.execute("INSERT INTO fileids VALUES (?, ?, ?)", (idx, path, mtime))
                await db.commit()
                return idx

    async def watch_files(self):
        async with self.lock:
            async with aiosqlite.connect(self.db_path) as db:
                await db.execute("DROP TABLE IF EXISTS fileids")
                await db.execute(
                    "CREATE TABLE fileids "
                    "(id TEXT PRIMARY KEY, path TEXT NOT NULL UNIQUE, mtime REAL NOT NULL)"
                )
                await db.commit()

        # index files
        async with self.lock:
            async with aiosqlite.connect(self.db_path) as db:
                async for path in Path().rglob("*"):
                    idx = uuid4().hex
                    mtime = (await path.stat()).st_mtime
                    await db.execute(
                        "INSERT INTO fileids VALUES (?, ?, ?)", (idx, str(path), mtime)
                    )
                await db.commit()
                self.initialized.set()

        async for changes in awatch(".", stop_event=self.stop_watching_files):
            async with self.lock:
                async with aiosqlite.connect(self.db_path) as db:
                    deleted_paths = []
                    added_paths = []
                    for change, changed_path in changes:
                        # get relative path
                        changed_path = Path(changed_path).relative_to(await Path().absolute())
                        changed_path_str = str(changed_path)

                        if change == Change.deleted:
                            logger.debug("File %s was deleted", changed_path_str)
                            async with db.execute(
                                "SELECT COUNT(*) FROM fileids WHERE path = ?", (changed_path_str,)
                            ) as cursor:
                                if not (await cursor.fetchone())[0]:
                                    # path is not indexed, ignore
                                    logger.debug(
                                        "File %s is not indexed, ignoring", changed_path_str
                                    )
                                    continue
                            # path is indexed
                            await maybe_rename(
                                db, changed_path_str, deleted_paths, added_paths, False
                            )
                        elif change == Change.added:
                            logger.debug("File %s was added", changed_path_str)
                            await maybe_rename(
                                db, changed_path_str, added_paths, deleted_paths, True
                            )
                        elif change == Change.modified:
                            logger.debug("File %s was modified", changed_path_str)
                            if changed_path_str == self.db_path:
                                continue
                            async with db.execute(
                                "SELECT COUNT(*) FROM fileids WHERE path = ?", (changed_path_str,)
                            ) as cursor:
                                if not (await cursor.fetchone())[0]:
                                    # path is not indexed, ignore
                                    logger.debug(
                                        "File %s is not indexed, ignoring", changed_path_str
                                    )
                                    continue
                            mtime = (await changed_path.stat()).st_mtime
                            await db.execute(
                                "UPDATE fileids SET mtime = ? WHERE path = ?",
                                (mtime, changed_path_str),
                            )

                    for path in deleted_paths + added_paths:
                        await db.execute("DELETE FROM fileids WHERE path = ?", (path,))
                    await db.commit()

            for change in changes:
                changed_path = change[1]
                # get relative path
                changed_path = str(Path(changed_path).relative_to(await Path().absolute()))
                for watcher in self.watchers.get(changed_path, []):
                    watcher.notify(change)

        self.stopped_watching_files.set()

    def watch(self, path: str) -> Watcher:
        watcher = Watcher(path)
        self.watchers.setdefault(path, []).append(watcher)
        return watcher

    def unwatch(self, path: str, watcher: Watcher):
        self.watchers[path].remove(watcher)


async def get_mtime(path, db) -> Optional[float]:
    if db:
        async with db.execute("SELECT mtime FROM fileids WHERE path = ?", (path,)) as cursor:
            async for mtime, in cursor:
                return mtime
            # deleted file is not in database, shouldn't happen
            return None
    try:
        mtime = (await Path(path).stat()).st_mtime
    except FileNotFoundError:
        return None
    return mtime


async def maybe_rename(
    db, changed_path: str, changed_paths: List[str], other_paths: List[str], is_added_path
) -> None:
    # check if the same file was added/deleted, this would be a rename
    db_or_fs1, db_or_fs2 = db, None
    if is_added_path:
        db_or_fs1, db_or_fs2 = db_or_fs2, db_or_fs1
    mtime1 = await get_mtime(changed_path, db_or_fs1)
    if mtime1 is None:
        return
    for other_path in other_paths:
        mtime2 = await get_mtime(other_path, db_or_fs2)
        if mtime1 == mtime2:
            # same files, according to modification times
            path1, path2 = changed_path, other_path
            if is_added_path:
                path1, path2 = path2, path1
            logger.debug("File %s was renamed to %s", path1, path2)
            await db.execute("UPDATE fileids SET path = ? WHERE path = ?", (path2, path1))
            other_paths.remove(other_path)
            return
    changed_paths.append(changed_path)



import asyncio
import os

from jupyverse_api.terminals import TerminalServer
from winpty import PTY  # type: ignore


def open_terminal(command="C:\\Windows\\System32\\cmd.exe", columns=80, lines=24):
    env = "\0".join([f"{k}={v}" for k, v in os.environ.items()]) + "\0"
    process = PTY(columns, lines)
    process.spawn(command, env=env)
    return process


class _TerminalServer(TerminalServer):
    def __init__(self):
        self.process = open_terminal()
        self.websockets = []

    async def serve(self, websocket):
        self.websocket = websocket
        self.websockets.append(websocket)

        await websocket.send_json(["setup", {}])

        self.send_task = asyncio.create_task(self.send_data())
        self.recv_task = asyncio.create_task(self.recv_data())

        await asyncio.gather(self.send_task, self.recv_task)

    async def send_data(self):
        while True:
            try:
                data = self.process.read(blocking=False)
            except Exception:
                await self.websocket.send_json(["disconnect", 1])
                return
            if not data:
                await asyncio.sleep(0.1)
            else:
                for websocket in self.websockets:
                    await websocket.send_json(["stdout", data])

    async def recv_data(self):
        while True:
            try:
                msg = await self.websocket.receive_json()
            except Exception:
                return
            if msg[0] == "stdin":
                self.process.write(msg[1])
            elif msg[0] == "set_size":
                self.process.set_size(msg[2], msg[1])

    def quit(self, websocket):
        self.websockets.remove(websocket)
        if not self.websockets:
            self.send_task.cancel()
            self.recv_task.cancel()
            del self.process



from typing import Any, Dict, List, Optional, Tuple

from fastapi import Depends, HTTPException, Request, Response, WebSocket, status
from fastapi.security import APIKeyCookie
from fief_client import FiefAccessTokenInfo, FiefAsync, FiefUserInfo  # type: ignore
from fief_client.integrations.fastapi import FiefAuth  # type: ignore
from jupyverse_api.auth import User

from .config import _AuthFiefConfig


class Backend:
    def __init__(self, auth_fief_config: _AuthFiefConfig):
        class CustomFiefAuth(FiefAuth):
            client: FiefAsync

            async def get_unauthorized_response(self, request: Request, response: Response):
                redirect_uri = str(request.url_for("auth_callback"))
                auth_url = await self.client.auth_url(redirect_uri, scope=["openid"])
                raise HTTPException(
                    status_code=status.HTTP_307_TEMPORARY_REDIRECT,
                    headers={"Location": auth_url},
                )

        self.fief = FiefAsync(
            auth_fief_config.base_url,
            auth_fief_config.client_id,
            auth_fief_config.client_secret,
        )

        self.SESSION_COOKIE_NAME = "fps_auth_fief_user_session"
        scheme = APIKeyCookie(name=self.SESSION_COOKIE_NAME, auto_error=False)
        self.auth = CustomFiefAuth(self.fief, scheme)

        async def update_user(
            user: FiefUserInfo = Depends(self.auth.current_user()),
            access_token_info: FiefAccessTokenInfo = Depends(self.auth.authenticated()),
        ):
            async def _(data: Dict[str, Any]) -> FiefUserInfo:
                user = await self.fief.update_profile(
                    access_token_info["access_token"], {"fields": data}
                )
                return user

            return _

        def websocket_auth(permissions: Optional[Dict[str, List[str]]] = None):
            async def _(
                websocket: WebSocket,
            ) -> Optional[Tuple[WebSocket, Optional[Dict[str, List[str]]]]]:
                accept_websocket = False
                checked_permissions: Optional[Dict[str, List[str]]] = None
                if self.SESSION_COOKIE_NAME in websocket._cookies:
                    access_token = websocket._cookies[self.SESSION_COOKIE_NAME]
                    if permissions is None:
                        accept_websocket = True
                    else:
                        checked_permissions = {}
                        for resource, actions in permissions.items():
                            allowed = checked_permissions[resource] = []
                            for action in actions:
                                try:
                                    await self.fief.validate_access_token(
                                        access_token, required_permissions=[f"{resource}:{action}"]
                                    )
                                except BaseException:
                                    pass
                                else:
                                    allowed.append(action)
                                    accept_websocket = True
                if accept_websocket:
                    return websocket, checked_permissions
                else:
                    await websocket.close(code=status.WS_1008_POLICY_VIOLATION)
                    return None

            return _

        def current_user(permissions=None):
            if permissions is not None:
                permissions = [
                    f"{resource}:{action}"
                    for resource, actions in permissions.items()
                    for action in actions
                ]

            async def _(
                user: FiefUserInfo = Depends(self.auth.current_user(permissions=permissions)),
            ):
                return User(**user["fields"])

            return _

        self.current_user = current_user
        self.update_user = update_user
        self.websocket_auth = websocket_auth



import json
from pathlib import Path

import retrolab  # type: ignore
from fastapi import APIRouter, Depends
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.lab import Lab
from jupyverse_api.retrolab import RetroLab


class _RetroLab(RetroLab):
    def __init__(self, app: App, auth: Auth, frontend_config: FrontendConfig, lab: Lab) -> None:
        super().__init__(app)

        router = APIRouter()
        prefix_dir, federated_extensions = lab.init_router(router, "retro/tree")
        retrolab_dir = Path(retrolab.__file__).parent

        self.mount(
            "/static/retro",
            StaticFiles(directory=retrolab_dir / "static"),
            name="static",
        )

        for path in (retrolab_dir / "labextension" / "static").glob("remoteEntry.*.js"):
            load = f"static/{path.name}"
            break
        retro_federated_extensions = [
            {
                "extension": "./extension",
                "load": load,
                "name": "@retrolab/lab-extension",
                "style": "./style",
            }
        ]

        @router.get("/retro/tree", response_class=HTMLResponse)
        async def get_tree(
            user: User = Depends(auth.current_user()),
        ):
            return get_index(
                "Tree", "tree", frontend_config.collaborative, frontend_config.base_url
            )

        @router.get("/retro/notebooks/{path:path}", response_class=HTMLResponse)
        async def get_notebook(
            path,
            user: User = Depends(auth.current_user()),
        ):
            return get_index(
                path, "notebooks", frontend_config.collaborative, frontend_config.base_url
            )

        @router.get("/retro/edit/{path:path}", response_class=HTMLResponse)
        async def edit_file(
            path,
            user: User = Depends(auth.current_user()),
        ):
            return get_index(path, "edit", frontend_config.collaborative, frontend_config.base_url)

        @router.get("/retro/consoles/{path:path}", response_class=HTMLResponse)
        async def get_console(
            path,
            user: User = Depends(auth.current_user()),
        ):
            return get_index(
                path, "consoles", frontend_config.collaborative, frontend_config.base_url
            )

        @router.get("/retro/terminals/{name}", response_class=HTMLResponse)
        async def get_terminal(
            name: str,
            user: User = Depends(auth.current_user()),
        ):
            return get_index(
                name, "terminals", frontend_config.collaborative, frontend_config.base_url
            )

        def get_index(doc_name, retro_page, collaborative, base_url="/"):
            extensions_dir = prefix_dir / "share" / "jupyter" / "labextensions"
            federated_extensions, disabled_extension = lab.get_federated_extensions(extensions_dir)
            page_config = {
                "appName": "RetroLab",
                "appNamespace": "retro",
                "appSettingsDir": (
                    prefix_dir / "share" / "jupyter" / "lab" / "settings"
                ).as_posix(),
                "appUrl": "/lab",
                "appVersion": retrolab.__version__,
                "baseUrl": base_url,
                "cacheFiles": True,
                "collaborative": collaborative,
                "disabledExtensions": disabled_extension,
                "extraLabextensionsPath": [],
                "federated_extensions": retro_federated_extensions + federated_extensions,
                "frontendUrl": "/retro/",
                "fullAppUrl": f"{base_url}lab",
                "fullLabextensionsUrl": f"{base_url}lab/extensions",
                "fullLicensesUrl": f"{base_url}lab/api/licenses",
                "fullListingsUrl": f"{base_url}lab/api/listings",
                "fullMathjaxUrl": f"{base_url}static/notebook/components/MathJax/MathJax.js",
                "fullSettingsUrl": f"{base_url}lab/api/settings",
                "fullStaticUrl": f"{base_url}static/retro",
                "fullThemesUrl": f"{base_url}lab/api/themes",
                "fullTranslationsApiUrl": f"{base_url}lab/api/translations",
                "fullTreeUrl": f"{base_url}lab/tree",
                "fullWorkspacesApiUrl": f"{base_url}lab/api/workspaces",
                "labextensionsPath": [
                    (prefix_dir / "share" / "jupyter" / "labextensions").as_posix()
                ],
                "labextensionsUrl": "/lab/extensions",
                "licensesUrl": "/lab/api/licenses",
                "listingsUrl": "/lab/api/listings",
                "mathjaxConfig": "TeX-AMS-MML_HTMLorMML-full,Safe",
                "retroLogo": False,
                "retroPage": retro_page,
                "schemasDir": (prefix_dir / "share" / "jupyter" / "lab" / "schemas").as_posix(),
                "settingsUrl": "/lab/api/settings",
                "staticDir": (retrolab_dir / "static").as_posix(),
                "templatesDir": (retrolab_dir / "templates").as_posix(),
                "terminalsAvailable": True,
                "themesDir": (prefix_dir / "share" / "jupyter" / "lab" / "themes").as_posix(),
                "themesUrl": "/lab/api/themes",
                "translationsApiUrl": "/lab/api/translations",
                "treeUrl": "/lab/tree",
                "workspacesApiUrl": "/lab/api/workspaces",
                "wsUrl": "",
            }
            index = (
                INDEX_HTML.replace("PAGE_CONFIG", json.dumps(page_config))
                .replace("DOC_NAME", doc_name)
                .replace("BASE_URL", base_url)
            )
            return index

        self.include_router(router)


INDEX_HTML = """\
<!DOCTYPE html>
<html>
<head>
  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <title>RetroLab - DOC_NAME</title>
  <link rel="icon" type="image/x-icon" href="/static/favicons/favicon-notebook.ico"
    class="favicon">
</head>
<body>
  <script id="jupyter-config-data" type="application/json">
    PAGE_CONFIG
  </script>
  <script src="BASE_URLstatic/retro/bundle.js" main="index"></script>
  <script type="text/javascript">
    /* Remove token from URL. */
    (function () {
      var parsedUrl = new URL(window.location.href);
      if (parsedUrl.searchParams.get('token')) {
        parsedUrl.searchParams.delete('token');
        window.history.replaceState({ }, '', parsedUrl.href);
      }
    })();
  </script>
</body>
</html>
"""



from __future__ import annotations

import webbrowser
from typing import Any, Callable, Dict, Sequence

from asgiref.typing import ASGI3Application
from asphalt.core import Component, Context
from asphalt.web.fastapi import FastAPIComponent
from fastapi import FastAPI
from pydantic import BaseModel

from ..app import App


class AppComponent(Component):
    async def start(
        self,
        ctx: Context,
    ) -> None:
        app = await ctx.request_resource(FastAPI)

        _app = App(app)
        ctx.add_resource(_app)


class JupyverseComponent(FastAPIComponent):
    def __init__(
        self,
        components: dict[str, dict[str, Any] | None] | None = None,
        *,
        app: FastAPI | str | None = None,
        host: str = "127.0.0.1",
        port: int = 8000,
        open_browser: bool = False,
        query_params: Dict[str, Any] | None = None,
        debug: bool | None = None,
        middlewares: Sequence[Callable[..., ASGI3Application] | dict[str, Any]] = (),
    ) -> None:
        super().__init__(
            components,  # type: ignore
            app=app,
            host=host,
            port=port,
            debug=debug,
            middlewares=middlewares,
        )
        self.host = host
        self.port = port
        self.open_browser = open_browser
        self.query_params = query_params

    async def start(
        self,
        ctx: Context,
    ) -> None:
        query_params = QueryParams(d={})
        host = self.host
        if not host.startswith("http"):
            host = f"http://{host}"
        host_url = Host(url=f"{host}:{self.port}/")
        ctx.add_resource(query_params)
        ctx.add_resource(host_url)

        await super().start(ctx)

        # at this point, the server has started
        if self.open_browser:
            qp = query_params.d
            if self.query_params:
                qp.update(**self.query_params)
            query_params_str = "?" + "&".join([f"{k}={v}" for k, v in qp.items()]) if qp else ""
            webbrowser.open_new_tab(f"{self.host}:{self.port}{query_params_str}")


class QueryParams(BaseModel):
    d: Dict[str, Any]


class Host(BaseModel):
    url: str



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from jupyverse_api import Router, Config


class Kernels(Router, ABC):
    @abstractmethod
    async def watch_connection_files(self, path: Path) -> None:
        ...


class KernelsConfig(Config):
    default_kernel: str = "python3"
    connection_path: Optional[str] = None



import asyncio
import fcntl
import os
import pty
import shlex
import struct
import termios

from fastapi import WebSocketDisconnect
from jupyverse_api.terminals import TerminalServer


def open_terminal(command="bash", columns=80, lines=24):
    pid, fd = pty.fork()
    if pid == 0:
        argv = shlex.split(command)
        env = os.environ.copy()
        env.update(TERM="linux", COLUMNS=str(columns), LINES=str(lines))
        os.execvpe(argv[0], argv, env)
    return fd


class _TerminalServer(TerminalServer):
    def __init__(self):
        self.fd = open_terminal()
        self.p_out = os.fdopen(self.fd, "w+b", 0)
        self.websockets = []

    async def serve(self, websocket, permissions):
        self.websocket = websocket
        self.websockets.append(websocket)
        self.event = asyncio.Event()
        self.loop = asyncio.get_event_loop()

        task = asyncio.create_task(self.send_data())

        def on_output():
            try:
                self.data_or_disconnect = self.p_out.read(65536).decode()
                self.event.set()
            except Exception:
                self.loop.remove_reader(self.p_out)
                self.data_or_disconnect = None
                self.event.set()

        self.loop.add_reader(self.p_out, on_output)
        await websocket.send_json(["setup", {}])
        can_execute = permissions is None or "execute" in permissions.get("terminals", [])
        try:
            while True:
                msg = await websocket.receive_json()
                if can_execute:
                    if msg[0] == "stdin":
                        self.p_out.write(msg[1].encode())
                    elif msg[0] == "set_size":
                        winsize = struct.pack("HH", msg[1], msg[2])
                        fcntl.ioctl(self.fd, termios.TIOCSWINSZ, winsize)
        except WebSocketDisconnect:
            task.cancel()

    async def send_data(self):
        while True:
            await self.event.wait()
            self.event.clear()
            if self.data_or_disconnect is None:
                await self.websocket.send_json(["disconnect", 1])
            else:
                for websocket in self.websockets:
                    await websocket.send_json(["stdout", self.data_or_disconnect])

    def quit(self, websocket):
        self.websockets.remove(websocket)
        if not self.websockets:
            os.close(self.fd)



from datetime import datetime
from http import HTTPStatus
from typing import Any, Dict, Type

from fastapi import APIRouter, Depends, Response
from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.terminals import Terminals, TerminalServer

from .models import Terminal

TERMINALS: Dict[str, Dict[str, Any]] = {}


class _Terminals(Terminals):
    def __init__(self, app: App, auth: Auth, _TerminalServer: Type[TerminalServer]) -> None:
        super().__init__(app)
        router = APIRouter()

        @router.get("/api/terminals")
        async def get_terminals(
            user: User = Depends(auth.current_user({"terminals": ["read"]})),
        ):
            return [terminal["info"] for terminal in TERMINALS.values()]

        @router.post("/api/terminals")
        async def create_terminal(
            user: User = Depends(auth.current_user({"terminals": ["write"]})),
        ):
            name = str(len(TERMINALS) + 1)
            terminal = Terminal(
                name=name,
                last_activity=datetime.utcnow().isoformat() + "Z",
            )
            server = _TerminalServer()
            TERMINALS[name] = {"info": terminal, "server": server}
            return terminal

        @router.delete("/api/terminals/{name}", status_code=204)
        async def delete_terminal(
            name: str,
            user: User = Depends(auth.current_user(permissions={"terminals": ["write"]})),
        ):
            for websocket in TERMINALS[name]["server"].websockets:
                TERMINALS[name]["server"].quit(websocket)
            del TERMINALS[name]
            return Response(status_code=HTTPStatus.NO_CONTENT.value)

        @router.websocket("/terminals/websocket/{name}")
        async def terminal_websocket(
            name,
            websocket_permissions=Depends(
                auth.websocket_auth(permissions={"terminals": ["read", "execute"]})
            ),
        ):
            if websocket_permissions is None:
                return
            websocket, permissions = websocket_permissions
            await websocket.accept()
            await TERMINALS[name]["server"].serve(websocket, permissions)
            if name in TERMINALS:
                TERMINALS[name]["server"].quit(websocket)
                if not TERMINALS[name]["server"].websockets:
                    del TERMINALS[name]

        self.include_router(router)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Tuple

from fastapi import APIRouter
from jupyverse_api import Router


class Lab(Router, ABC):
    @abstractmethod
    def init_router(
        self, router: APIRouter, redirect_after_root: str
    ) -> Tuple[Path, List[Dict[str, Any]]]:
        ...

    @abstractmethod
    def get_federated_extensions(self, extensions_dir: Path) -> Tuple[List, List]:
        ...



from asphalt.core import Component, Context
from jupyverse_api.frontend import FrontendConfig


class FrontendComponent(Component):
    def __init__(self, **kwargs):
        self.frontend_config = FrontendConfig(**kwargs)

    async def start(
        self,
        ctx: Context,
    ) -> None:
        ctx.add_resource(self.frontend_config, types=FrontendConfig)



import asyncio
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, Union

from jupyverse_api import Router

from .models import Content, SaveContent


class FileIdManager(ABC):
    stop_watching_files: asyncio.Event
    stopped_watching_files: asyncio.Event

    @abstractmethod
    async def get_path(self, file_id: str) -> str:
        ...

    @abstractmethod
    async def get_id(self, file_path: str) -> str:
        ...


class Contents(Router, ABC):
    @property
    @abstractmethod
    def file_id_manager(self) -> FileIdManager:
        ...

    @abstractmethod
    async def read_content(
        self, path: Union[str, Path], get_content: bool, as_json: bool = False
    ) -> Content:
        ...

    @abstractmethod
    async def write_content(self, content: Union[SaveContent, Dict]) -> None:
        ...



from typing import Dict, List, Optional, Union

from pydantic import BaseModel


class Content(BaseModel):
    name: str
    path: str
    last_modified: Optional[str]
    created: Optional[str]
    content: Optional[Union[str, Dict, List[Dict]]]
    format: Optional[str]
    mimetype: Optional[str]
    size: Optional[int]
    writable: bool
    type: str


class SaveContent(BaseModel):
    content: Optional[Union[str, Dict]]
    format: str
    path: str
    type: str



import psutil
from anyio import to_thread
from fastapi import APIRouter, Depends
from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.resource_usage import ResourceUsage, ResourceUsageConfig


class _ResourceUsage(ResourceUsage):
    def __init__(
        self,
        app: App,
        auth: Auth,
        resource_usage_config: ResourceUsageConfig,
    ):
        super().__init__(app)

        router = APIRouter()

        @router.get("/api/metrics/v1")
        async def get_metrics(
            user: User = Depends(auth.current_user(permissions={"contents": ["read"]})),
        ):
            cur_process = psutil.Process()
            all_processes = [cur_process] + cur_process.children(recursive=True)

            # Get memory information
            rss = 0
            for p in all_processes:
                try:
                    rss += p.memory_info().rss
                except (psutil.NoSuchProcess, psutil.AccessDenied):
                    pass

            mem_limit = resource_usage_config.mem_limit

            limits = {"memory": {"rss": mem_limit}}
            if resource_usage_config.mem_limit and resource_usage_config.mem_warning_threshold:
                limits["memory"]["warn"] = (mem_limit - rss) < (
                    mem_limit * resource_usage_config.mem_warning_threshold
                )

            metrics = {"rss": rss, "limits": limits}

            # Optionally get CPU information
            if resource_usage_config.track_cpu_percent:
                cpu_count = psutil.cpu_count()
                cpu_percent = await to_thread.run_sync(_get_cpu_percent, all_processes)

                if resource_usage_config.cpu_limit:
                    limits["cpu"] = {"cpu": resource_usage_config.cpu_limit}
                    if resource_usage_config.cpu_warning_threshold:
                        limits["cpu"]["warn"] = (resource_usage_config.cpu_limit - cpu_percent) < (
                            resource_usage_config.cpu_limit
                            * resource_usage_config.cpu_warning_threshold
                        )

                metrics.update(cpu_percent=cpu_percent, cpu_count=cpu_count)

            return metrics

        self.include_router(router)


def _get_cpu_percent(all_processes):
    def get_cpu_percent(p):
        try:
            return p.cpu_percent(interval=0.05)
        # Avoid littering logs with stack traces complaining
        # about dead processes having no CPU usage
        except BaseException:
            return 0

    return sum([get_cpu_percent(p) for p in all_processes])



from asphalt.core import Component, Context
from jupyverse_api.auth import Auth
from jupyverse_api.app import App
from jupyverse_api.resource_usage import ResourceUsage, ResourceUsageConfig

from .routes import _ResourceUsage


class ResourceUsageComponent(Component):
    def __init__(self, **kwargs):
        self.resource_usage_config = ResourceUsageConfig(**kwargs)

    async def start(
        self,
        ctx: Context,
    ) -> None:
        app = await ctx.request_resource(App)
        auth = await ctx.request_resource(Auth)  # type: ignore

        resource_usage = _ResourceUsage(app, auth, self.resource_usage_config)
        ctx.add_resource(resource_usage, types=ResourceUsage)



from jupyverse_api import Router, Config


class ResourceUsage(Router):
    pass


class ResourceUsageConfig(Config):
    mem_limit: int = 0
    mem_warning_threshold: int = 0
    track_cpu_percent: bool = False
    cpu_limit: int = 0
    cpu_warning_threshold: int = 0



from jupyverse_api.auth import AuthConfig


class _AuthFiefConfig(AuthConfig):
    base_url: str  # Base URL of Fief tenant
    client_id: str  # ID of Fief client
    client_secret: str  # Secret of Fief client



from typing import Optional

from pydantic import BaseModel


class Checkpoint(BaseModel):
    id: str
    last_modified: str


class CreateContent(BaseModel):
    ext: Optional[str]
    path: str
    type: str


class RenameContent(BaseModel):
    path: str
